# -*- test-case-name: morbid.tests.test_stompproto -*-
import traceback
import uuid
from datetime import datetime

from morbid.mqsecurity import Connector, IConnector, MQPortal
from morbid.restq import RestQ
from messagequeue import MessageQueueManager
from morbid.messagequeue import QueueError
from morbid.error import error
from collections import deque

verbose = False

try:
    from twisted.internet.protocol import Factory, Protocol
    from twisted.protocols.basic import LineReceiver
    from twisted.python import log
except ImportError:
    error("Twisted required to run; see www.twistedmatrix.com", False)


class StompProtocol(LineReceiver):
    delimiter = '\x00'
    __content_length = None
    __msg = None
    #TODO: we should override this with a config variable
    # this length measures the real size of the message including headers
    # this is a protection against attackers that would send thousands of
    # headers and saturate the server memory
    MAX_LENGTH = 16384
    # maximum number of frames a transaction is allowed
    MAX_TRANSACTION_SIZE = 1000
    # maximum transaction number a client is allowed to open at the same time
    MAX_CONCURRENT_TRANSACTIONS = 5

    def __init__(self, useRq=False):
        self.state = 'initial'
        self.username = None
        self.password = None
        self.buffer = ""
        self.id = str(uuid.uuid4())
        self.avatar = None
        self.receivedDisconnect = False
        self.useRq = useRq
        self.transactionBuffer = dict()

    def get_groups(self):
        return self.avatar.groups

    def lineLengthExceeded(self, line):
        """this is an override of the L{LineReceiver} implementation.
        If the line length exceeds the specified sized we send an error
        message and then drop the connection.
        """
        self.sendFrame(
            'ERROR',
            {'message': "Max frame len exceeded"},
            "Max frame len exceeded"
        )
        self.transport.loseConnection()

    def rawDataReceived(self, data):
        if '\x00' in data:
            datalist = data.split('\x00')

            for index, item in enumerate(datalist):
                self.__msg['body'] += item

                if len(self.__msg['body']) == self.__content_length:
                    self.read_connected(**self.__msg)
                    self.__msg = None
                    self.__content_length = None
                    self.setLineMode(extra='\x00'.join(datalist[index+1:]))
                    return
                else:
                    self.__msg['body'] += '\x00'

        else:
            self.__msg['body'] += data


    def lineReceived(self, line):
        # TODO: are we sure we want to support broken clients?
        # Stomper adds an \n after the \x00 at the end of the frame
        # which is clearly not in the protocol
        # this is corrupting the buffer for the next frame
        # for the moment we just accept this...
        # we encounter the problem only during the tests and not with a real
        # client. So as soon as we rewrite the tests to avoid stomper we
        # will be able to remove this hack!
        line = line.lstrip('\n')

        log.msg('message received on protocol id: %s' % self.id)
        try:
            hdata, body = line.split('\n\n', 1)
            elems = hdata.split('\n')

        except ValueError:
            # if we have a value error this must be the client
            # who corrupted the buffer
            msg = "Garbage in your last frame: '%s'" % repr(line)
            self.sendFrame('ERROR', {'message': msg}, msg)
            return

        cmd = elems.pop(0).lower().strip()

        headers = {}
        # We can't use a simple split because the value can legally contain
        # colon characters (for example, the session returned by ActiveMQ).
        for e in elems:
            try:
                i = e.find(':')
            except ValueError:
                continue

            k = e[:i].strip()
            v = e[i+1:].strip()
            headers[k] = v

        msg = {
            'cmd': cmd,
            'headers': headers,
            'body': body,
            }

        if 'content-length' in headers.keys():
            # see if we have the full body in our "line" or if we hit
            # a \x00 inside the body
            content_length = int(headers['content-length'])
            if not len(body) == content_length:
                # go to raw mode to manage the other packets
                self.__content_length = content_length
                msg['body'] += '\x00'
                self.__msg = msg
                self.setRawMode()
                return

        if self.state == 'initial':
            self.read_initial(**msg)

        elif self.state == 'connected':
            self.read_connected(**msg)

    def sendError(self, e):
        exception, instance, tb = traceback.sys.exc_info()
        tbOutput= "".join(traceback.format_tb(tb))
        self.sendFrame('ERROR', {'message': str(e) }, tbOutput)

    def sendFrame(self, cmd, headers, body):
        hdrs = '\n'.join(['%s:%s' % (k,v) for k,v in headers.items()])
        frame = '%s\n%s\n\n%s\x00' % (cmd, hdrs, body)
        self.transport.write(frame)

    def doReceipt(self, receiptId):
        """send a RECEIPT frame
        """
        cmd = 'RECEIPT'
        hdrs = 'receipt-id:%s' % receiptId
        frame = '%s\n%s\n\n\x00' % (cmd, hdrs)
        self.transport.write(frame)

    def autosubscribe(self, subs):
        for sub in subs:
            self.frame_subscribe(({'destination':sub}, None))

    def autounsubscribe(self, unsubs):
        for unsub in unsubs:
            self.frame_unsubscribe(({'destination':unsub}, None))

    def read_initial(self, cmd, headers, body):
        assert cmd == 'connect', "Invalid cmd: expected CONNECT"
        self.username = headers.get('login',"")
        self.password = headers.get('passcode',"")
        d = None

        if self.factory.mq_portal:
            d = self.factory.mq_portal.stomp_login(**headers)

        if d:
            d.addCallback(self.stomp_connected)
            d.addErrback(self.stomp_connect_failed)

        else:
            self.stomp_connected(
                (IConnector, Connector("None", ["None"]), lambda: None))

    def stomp_connected(self, *args):

        def stomp_connected_callback((headers, body)):
            if "allow" in headers and headers["allow"] == "no":
                return self.stomp_connect_failed()

            avatarInterface, avatar, logout_func = args[0]
            self.avatar = avatar
            self.state = 'connected'
            self.sendFrame('CONNECTED', {"session": self.id}, "")

        if self.useRq:
            d = self.factory.restq.submit(
                self, "connect",
                {"username":self.username, "password":self.password}
            )
            d.addCallback(stomp_connected_callback)

        else:
            avatarInterface, avatar, logout_func = args[0]
            self.avatar = avatar
            self.state = 'connected'
            self.sendFrame('CONNECTED', {"session": self.id}, "")

    def stomp_connect_failed(self, *args):
        self.sendFrame(
            'ERROR',
            {'message': "Invalid ID or password"},
            "Invalid ID or password"
        )
        self.username = None
        self.transport.loseConnection()

    def read_connected(self, cmd, headers, body):
        if headers.has_key('receipt'):
            # handle the receipt to our client
            self.doReceipt(headers['receipt'].strip())

        txid = headers.get('transaction', None)
        if txid:
            txid = txid.strip()

        if cmd == 'begin':
            # initialize a transaction buffer on this protocol using the
            # transaction unique identifier.
            if not txid or txid == '':
                # a BEGIN frame MUST have a transaction id
                self.sendFrame(
                    'ERROR',
                    {'message': "Missing transaction id"},
                    "Missing transaction id"
                )
                # TODO: should we lose the connection or just continue?
                self.transport.loseConnection()
            else:
                if len(self.transactionBuffer.keys()) >= self.MAX_CONCURRENT_TRANSACTIONS:
                    self.sendFrame(
                        'ERROR',
                            {'message': "Cannot create more transactions for you"},
                        "Cannot create more transactions for you"
                    )
                    return

                self.transactionBuffer[txid] = deque()
                return

        elif cmd == 'commit':
            # a COMMIT frame MUST have a transaction id
            if not txid or txid == '':
                # a BEGIN frame MUST have a transaction id
                self.sendFrame(
                    'ERROR',
                    {'message': "Missing transaction id"},
                    "Missing transaction id"
                )
                # TODO: should we lose the connection or just continue?
                self.transport.loseConnection()
                return

            else:
                # here we flush the queue en-masse
                q = self.transactionBuffer.pop(txid, None)
                if q is None:
                    self.sendFrame(
                        'ERROR',
                        {'message': "No such transaction: %s" % txid},
                        "No such transaction: %s" % txid
                    )
                    # TODO: should we lose the connection instead?
                    return

                self.__flush_transaction(q)
                return

        elif cmd == 'abort':
            # a ABORT frame MUST have a transaction id
            if not txid or txid == '':
                self.sendFrame(
                    'ERROR',
                        {'message': "Missing transaction id"},
                    "Missing transaction id"
                )
                # TODO: should we lose the connection instead?
                return

            q = self.transactionBuffer.pop(txid, None)
            if q is None:
                self.sendFrame(
                    'ERROR',
                        {'message': "No such transaction: %s" % txid},
                    "No such transaction: %s" % txid
                )

            # here we finished our job...
            return

        elif cmd == 'ack' and headers.get('transaction', None) is not None:
            if self.transactionBuffer.get(txid, None) is not None:
                if len(self.transactionBuffer[txid]) >= self.MAX_TRANSACTION_SIZE:
                    self.__max_transaction_reached(txid)
                    return

                self.transactionBuffer[txid].append((cmd, headers, body))
            return

        elif cmd == 'send' and headers.get('transaction', None) is not None:
            if self.transactionBuffer.get(txid, None) is not None:
                if len(self.transactionBuffer[txid]) >= self.MAX_TRANSACTION_SIZE:
                    self.__max_transaction_reached(txid)
                    return

                self.transactionBuffer[txid].append((cmd, headers, body))
            return

        self.__do_read_connected(cmd, headers, body)

    def __max_transaction_reached(self, txid):
        self.sendFrame(
            'ERROR',
                {'message': "Maximum size for this transaction reached"},
            "Maximum size for this transaction reached"
        )
        self.transactionBuffer.pop(txid, None)
        self.transport.loseConnection()

    def __do_read_connected(self, cmd, headers, body):
        if self.useRq:
            d = self.factory.restq.submit(self, cmd, headers, body)
            d.addCallback(getattr(self, 'frame_%s' % cmd))
        else:
            getattr(self, 'frame_%s' % cmd)((headers, body))

    def frame_subscribe(self, (headers, body)):
        """handle a SUBSCRIBE request from the connected client"""
        log.msg('client trying to subscribe to %s' % headers.get('destination', None))
        if "allow" in headers and headers["allow"] == "no":
            return

        if not headers.has_key('destination'):
            # the destination was not specified in the SUBSCRIBE frame!
            self.sendFrame(
                'ERROR',
                {'message': 'destination missing'},
                'destination missing')
            return

        try:
            self.factory.mqm.subscribe_queue(self, headers['destination'])

        except QueueError, err:
            self.sendFrame('ERROR',
                           {'message': self.get_message_code(err.code)},
                           self.get_message_text(err.code))

    def frame_unsubscribe(self, (headers, body)):
        self.factory.mqm.leave_queue(self, headers['destination'])

    def frame_send(self, (headers, body)):
        if "allow" in headers and headers["allow"] == "no":
            return
        try:
            result = self.factory.mqm.send_message(self, headers['destination'], (headers, body))
        except QueueError, err:
            self.sendFrame('ERROR',
                           {'message': self.get_message_code(err.code)},
                           self.get_message_text(err.code))

    def frame_disconnect(self, (headers, body)):
        self.receivedDisconnect = True
        self.transport.loseConnection()

    def connectionLost(self, reason):
        if not self.receivedDisconnect and self.username and self.useRq:
            self.factory.restq.submit(self, "disconnect")

        self.factory.disconnected(self)

    def get_message_code(self, code):
        return {'FAILC': "CREATE error",
                'FAILR': "READ error",
                'FAILW': "WRITE error"}[code]

    def get_message_text(self, code):
        return {'FAILC': "Not authorized to create queue",
                'FAILR': "Not authorized to read queue",
                'FAILW': "Not authorized to write to queue"}[code]

    def send(self, message):
        '''
        This method is invoked by the message queues.
        Not intended for direct use by the protocol.
        '''
        headers, body = message
        self.sendFrame('MESSAGE', headers, body)

    def __flush_transaction(self, q):
        """flush the whole queue since we received the COMMIT message
        """
        for cmd, headers, body in q:
            self.__do_read_connected(cmd, headers, body)


class StompFactory(Factory):
    """
    The StompFactory creates an instance of a StompProtocol for each connection.
    Successful authentication results in the creation of an avatar for that user.
    The Avatar is assigned to the StompProtocol.
    """

    def __init__(self, mqm=None, filename=None, rqaddr=None, verbose=False):
        self.id = 0
        if rqaddr:
            self.restq = RestQ(rqaddr)
        else:
            self.restq = None

        self.verbose = verbose
        if mqm:
            self.mqm = mqm
        else:
            self.mqm = MessageQueueManager()
            # for the moment all queue names are open to any authenticated user
            self.mqm.queue_rights = lambda groups, queue: set(['c','r','w'])

        # TODO: we should also remove the filename arg and create the parms here
        self.mq_portal = MQPortal(filename=filename)

    def buildProtocol(self, addr):
        if self.restq:
            protocol = StompProtocol(useRq=True)
        else:
            protocol = StompProtocol(useRq=False)

        protocol.factory = self
        return protocol

    def report(self, msg):
        if self.verbose:
            print "[%s] MorbidQ: %s"%(datetime.now(), msg)

    def disconnected(self, proto):
        self.mqm.unsubscribe_all_queues(proto)

def load_config(config_file, config_dict):
    import os
    if not os.path.isfile(config_file):
        return error('config file "%s" does not exist'%(config_file,))
    f = open(config_file)
    lines = f.readlines()
    f.close()
    for line in lines:
        if "=" not in line:
            return error('error parsing config. line: "%s"'%(line,))
        option, value = line.split("=",1)
        if option == "verbose":
            if value.lower() == "true":
                value = True
            else:
                value = False
        if option == "port":
            try:
                value = int(value)
            except:
                return error('non-integer port specified: "%s"'%(line,))
        config_dict[option.strip()] = value.strip()

def get_stomp_factory(config_file="", cfg={'auth':None,'restq':None,'verbose':False}):
    if config_file:
        load_config(config_file, cfg)
    f = StompFactory(filename=cfg['auth'], rqaddr=cfg['restq'], verbose=cfg['verbose'])
    f.report("server starting with configuration: %s"%(cfg,))
    return f

def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option(
        "-c",
        "--config",
        dest="config",
        type="string",
        default="",
        help="configuration file"
    )
    parser.add_option(
        "-p",
        "--port",
        dest="port",
        type="int",
        default=61613,
        help="port the daemon should bind to (default: 61613)"
    )
    parser.add_option(
        "-i",
        "--interface",
        dest="interface",
        type="string",
        default="",
        help="hostname the daemon should bind to (default: all interfaces)"
    )
    parser.add_option(
        "-a",
        "--auth",
        dest="auth",
        type="string",
        default="",
        help="local path to authentication module"
    )
    parser.add_option(
        "-r",
        "--restq",
        dest="restq",
        type="string",
        default=None,
        help="RestQ http callback resource"
    )
    parser.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="verbose mode"
    )
    from twisted.internet import reactor
    (options, args) = parser.parse_args()
    stomp_factory = get_stomp_factory(options.config, {
        'verbose':options.verbose,
        'restq':options.restq,
        'auth':options.auth
    })
    reactor.listenTCP(options.port, stomp_factory, interface=options.interface)
    reactor.run()

if __name__ == "__main__":
    main()
