__author__ = 'faide'


from twisted.trial import unittest
from twisted.test import proto_helpers
from morbid.stompproto import StompFactory
from cStringIO import StringIO
import uuid


class StompProtoTestCase(unittest.TestCase):

    def setUp(self):
        factory = StompFactory()
        self.proto = factory.buildProtocol(('127.0.0.1', 0))
        self.tr = proto_helpers.StringTransport()
        self.proto.makeConnection(self.tr)

    def _send(self, command, headers, body):
        data = ''
        data += command + '\n'

        for k,v in headers.items():
            data += '%s:%s\n' % (k,v)

        # end of headers add another \n
        data += '\n'
        data += body
        # end of data marker
        data += '\x00'
        self.proto.dataReceived(data)

    def clean_buffer(self):
        # re-initialize the output buffer to a clean one because otherwise
        # the connection answer stays in the transport buffer
        self.tr.io = StringIO()

    def connect(self, cleanbuffer=False, receiptId=None):
        """a helper function to connect to the protocol
        """
        command = 'CONNECT'
        headers = {}
        body = ''
        if receiptId:
            headers['receipt'] = receiptId

        self._send(command, headers, body)

        if cleanbuffer:
            self.clean_buffer()

    def test_CONNECT(self):
        """test that the CONNECT command works as intended in the protocol...
        ie: that we receive a CONNECTED frame with a session id...
        """
        self.connect()
        expected = 'CONNECTED\nsession:%s\n\n\x00' % self.proto.id
        self.assertEquals(self.tr.value(), expected)

    def test_CONNECT_columninheader(self):
        """columns in the headers should be accepted by the protocol without
        issues. This test has nothing to do with login.
        """
        command = 'CONNECT'
        headers = {'login': 'super:man', 'passcode': 'super:pass'}
        body = ''
        self._send(command, headers, body)
        expected = 'CONNECTED\nsession:%s\n\n\x00' % self.proto.id
        self.assertEquals(self.tr.value(), expected)


    def test_SUBSCRIBE_without_destination(self):
        """the server should answer us about a missing destination
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        # here we omit to give a destination ...
        # this should prompt an ERROR reply
        headers = dict()
        body = ''
        self._send(command, headers, body)
        expected = 'ERROR\nmessage:destination missing\n\ndestination missing\x00'
        self.assertEquals(self.tr.value(), expected)

    def test_SUBSCRIBE(self):
        """make sure a valid SUBSCRIBE request is answered politely
        by the protocol
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        # this time we give a destination
        headers = dict(destination='somewhere')
        body = ''
        self._send(command, headers, body)
        # nothing should be answered by the server
        expected = ''
        self.assertEquals(self.tr.value(), expected)

    def test_SUBSCRIBED_RECEIVE_onSend(self):
        """we should receive our own message if we send it to a queue we
        are subscribed to.
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        # this time we give a destination
        headers = dict(destination='somewhere')
        body = ''
        self._send(command, headers, body)

        command = 'SEND'
        headers = dict(destination='somewhere')
        body = 'This is a message for the <somewhere> queue listeners'
        self._send(command, headers, body)

        # we should receive the message we just sent
        expected = 'MESSAGE\nmessage-id:somewhere_1\ndestination:somewhere\n\n'
        expected += 'This is a message for the <somewhere> '
        expected += 'queue listeners\x00'
        self.assertEquals(self.tr.value(), expected)

    def test_NONSUBSCRIBED_NORECEIVE_onSend(self):
        """we should not receive our own message if we send it to a queue we
        are not subscribed to.
        """
        self.connect(cleanbuffer=True)

        command = 'SEND'
        headers = dict(destination='somewhere')
        body = 'This is a message for the <somewhere> queue listeners'
        self._send(command, headers, body)

        # we should not receive the message we just sent
        expected = ''
        self.assertEquals(self.tr.value(), expected)

    def test_delimiter_in_messagebody_with_content_length(self):
        """in the spec it is said that if we want to send a \x00 in the message
        body we just have to give the content-lenght header to indicate the
        real body lenght...
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        headers = dict(destination='somewhere')
        body = ''
        self._send(command, headers, body)

        command = 'SEND'
        body = 'This is a message for \x00 you \x00 my little tester'
        headers = {'destination':'somewhere', 'content-length':str(len(body))}
        self._send(command, headers, body)

        # we should receive the message we just sent
        expected = 'MESSAGE\nmessage-id:somewhere_1\ncontent-length:%s\ndestination:somewhere\n\n' % len(body)
        expected += 'This is a message for \x00 you \x00 my little tester\x00'
        self.assertEquals(self.tr.value(), expected)

    def test_delimiter_in_messagebody_without_content_length(self):
        """in the spec it is said that if we want to send a \x00 in the message
        body we just have to give the content-lenght header to indicate the
        real body lenght...
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        headers = dict(destination='somewhere')
        body = ''
        self._send(command, headers, body)

        command = 'SEND'
        body = 'This is a message for \x00 you \x00 my little tester'
        headers = {'destination':'somewhere'}
        self._send(command, headers, body)

        # we should receive the message we just sent
        expected = "MESSAGE\nmessage-id:somewhere_1\ndestination:somewhere\n\n"
        expected += "This is a message for \x00"
        expected += "ERROR\nmessage:Garbage in your last frame: '' you ''\n\n"
        expected += "Garbage in your last frame: '' you ''\x00"
        expected += "ERROR\nmessage:Garbage in your last frame: '' my little tester''\n\n"
        expected += "Garbage in your last frame: '' my little tester''\x00"
        self.assertEquals(self.tr.value(), expected)

    def test_receipt_onSUBSCRIBE(self):
        """the protocol specs says that if the client gives a receipt header
        the server should send a RECEIPT response with the content
        of said header
        """
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        # this time we give a destination
        mId = str(uuid.uuid4())
        headers = dict(destination='somewhere', receipt=mId)
        body = ''
        self._send(command, headers, body)

        # we should receive an acknoledgement from the server in the
        # form of a receipt message with our receipt id we just sent
        expected = 'RECEIPT\nreceipt-id:%s\n\n\x00' % mId
        self.assertEquals(self.tr.value(), expected)

    def test_receipt_onCONNECT(self):
        """the protocol spec says that if the client gives a receipt header
        on the CONNECT command the server will not answer him...
        """
        mId = str(uuid.uuid4())
        self.connect(receiptId=mId)

        # the server should not give a RECEIPT on a connect command
        expected = 'CONNECTED\nsession:%s\n\n\x00' % self.proto.id
        self.assertEquals(self.tr.value(), expected)

    def test_transaction_commit_pushes_messages_in_queues(self):
        """test that messages in a transaction are sent only on the commit phase
        """

        # first we subscribe to a channel
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        headers = dict(destination='radio1')
        body = ''
        self._send(command, headers, body)

        command = 'BEGIN'
        headers = dict(transaction='trans1')
        body = ''
        self._send(command, headers, body)

        command = 'SEND'
        # this message is sent inside a transaction
        headers = dict(transaction='trans1', destination='radio1')
        body = "Radio 1, you're the only one!"
        self._send(command, headers, body)

        expected = ''
        self.assertEquals(self.tr.value(), expected)

        command = 'COMMIT'
        headers = dict(transaction='trans1')
        body = ''
        self._send(command, headers, body)

        expected = "MESSAGE\nmessage-id:radio1_1\n"
        expected += "destination:radio1\ntransaction:trans1\n\n"
        expected += "Radio 1, you're the only one!\x00"
        self.assertEquals(self.tr.value(), expected)

    def test_transaction_abort_loses_all_messages_in_queue(self):
        """test that messages in a transaction are not sent when the client
        aborts the said connection
        """

        # first we subscribe to a channel
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        headers = dict(destination='radio1')
        body = ''
        self._send(command, headers, body)

        command = 'BEGIN'
        headers = dict(transaction='trans1')
        body = ''
        self._send(command, headers, body)

        command = 'SEND'
        # this message is sent inside a transaction
        headers = dict(transaction='trans1', destination='radio1')
        body = "Radio 1, you're the only one!"
        self._send(command, headers, body)

        expected = ''
        self.assertEquals(self.tr.value(), expected)

        command = 'ABORT'
        headers = dict(transaction='trans1')
        body = ''
        self._send(command, headers, body)

        expected = ""
        self.assertEquals(self.tr.value(), expected)

    def test_max_transaction_size_reached(self):
        """if we reach the maximum size for a transaction we should receive
        and error and be disconnected
        """

        # set the transaction size to 3 for testing purposes
        self.proto.MAX_TRANSACTION_SIZE = 3

        # first we subscribe to a channel
        self.connect(cleanbuffer=True)
        command = 'SUBSCRIBE'
        headers = dict(destination='radio1')
        body = ''
        self._send(command, headers, body)

        command = 'BEGIN'
        headers = dict(transaction='trans1')
        body = ''
        self._send(command, headers, body)

        for i in [1, 2, 3, 4]:
            command = 'SEND'
            # this message is sent inside a transaction
            headers = dict(transaction='trans1', destination='radio1')
            body = "Radio 1, you're the only one!"
            self._send(command, headers, body)

        expected = 'ERROR\n'
        expected += 'message:Maximum size for this transaction reached\n\n'
        expected += 'Maximum size for this transaction reached\x00'
        self.assertEquals(self.tr.value(), expected)

        self.assertEquals(self.tr.disconnecting, True)
